"""Un/subscribe and consume AutoKitteh connection events."""

from datetime import timedelta
from uuid import uuid4

from .attr_dict import AttrDict


def subscribe(source: str, filter: str = "") -> str:
    """Subscribe to events on connection. Optional filter is a CEL expression."""
    # Dummy implementation for local development.
    return f"sig_{uuid4().hex}"


def unsubscribe(subscription_id: str) -> None:
    """Unsubscribe from events."""
    # Dummy implementation for local development.
    pass


def next_event(
    subscription_id: str | list[str], *, timeout: timedelta | int | float = None
) -> AttrDict:
    """Get the next event from the subscription(s).

    If timeout is not None and there are no new events after timeout, this function will
    return None.
    """
    # Dummy implementation for local development.
    return AttrDict()


def start(loc: str, data: dict = None, memo: dict = None) -> str:
    """Start a new session."""
    # Dummy implementation for local development.
    return f"ses_{uuid4().hex}"



import re
import sys
from subprocess import run


def _package_name(specifier):
    """Extract package name from package specifier.

    >>> _package_name('requests')
    'requests'
    >>> _package_name('requests==2.32.2')
    'requests'
    >>> _package_name('requests ~= 2.32')
    'requests'
    """

    match = re.match(r"\w+", specifier)
    if not match:
        raise ValueError(f"bad specifier: {specifier}")

    return match.group()


def _install_package(specifier, import_name):
    try:
        __import__(import_name)
        return
    except ImportError:
        pass

    out = run([sys.executable, "-m", "pip", "install", specifier])
    if out.returncode != 0:
        raise RuntimeError(f"can't install {specifier!r}")

    try:
        __import__(import_name)
    except ImportError:
        raise RuntimeError(
            f"can't import {import_name!r} after installing {specifier!r}"
        )


def install(*packages):
    """Install Python packages using pip.

    A package can be either a package requirement specifier
    (see https://pip.pypa.io/en/stable/reference/requirement-specifiers/)
    or a tuple of (package specifier, import name) in case the import name differs from
    the package name (e.g. Package `pillow` import imported as `PIL`).

    Examples:
    >>> install('requests', 'numpy')
    >>> install('requests ~= 2.32', 'numpy == 2.0.0')
    >>> install(['pillow ~= 10.4', 'PIL'])
    """
    for package in packages:
        if isinstance(package, str):
            specifier = package
            import_name = _package_name(specifier)
        elif isinstance(package, (tuple, list)):
            if len(package) != 2:
                raise ValueError(f"length should be 2: {package}")
            specifier, import_name = package
        else:
            raise TypeError(f"bad package: {package}")

        _install_package(specifier, import_name)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from pathlib import Path
import sys
import tomllib

sdk_dir = Path(__file__).parent.parent.absolute()
sys.path.insert(0, str(sdk_dir))  # Add the project root to the path

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "autokitteh"
copyright = "2024, AutoKitteh"
author = "AutoKitteh"
project_file = sdk_dir / "pyproject.toml"
with project_file.open("rb") as fp:
    data = tomllib.load(fp)
    release = data["project"]["version"]


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.napoleon"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]



"""Initialize a Twilio client, based on an AutoKitteh connection."""

import os

from twilio.rest import Client

from .connections import check_connection_name
from .errors import ConnectionInitError


def twilio_client(connection: str) -> Client:
    """Initialize a Twilio client, based on an AutoKitteh connection.

    API reference:
    https://www.twilio.com/docs/libraries/python

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Twilio client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
    """
    check_connection_name(connection)

    account_sid = os.getenv(connection + "__AccountSID")
    username = os.getenv(connection + "__Username")
    password = os.getenv(connection + "__Password")
    if not username or not password:
        raise ConnectionInitError(connection)

    return Client(username, password, account_sid=account_sid)



"""Unit tests for the "slack" module."""

from autokitteh import slack


def test_normalize_channel_name():
    assert slack.normalize_channel_name("") == ""
    assert slack.normalize_channel_name('"isn\'t"') == "isnt"
    assert slack.normalize_channel_name("TEST") == "test"
    assert slack.normalize_channel_name("1  2--3__4") == "1-2-3-4"
    assert slack.normalize_channel_name("a `~!@#$%^&*() 1") == "a-1"
    assert slack.normalize_channel_name("b -_=+ []{}|\\ 2") == "b-2"
    assert slack.normalize_channel_name("c ;:'\" ,.<>/? 3") == "c-3"
    assert slack.normalize_channel_name("-foo ") == "foo"



"""AutoKitteh connection-related utilities."""

from datetime import UTC, datetime
import re


def check_connection_name(connection: str) -> None:
    """Check that the given AutoKitteh connection name is valid.

    Args:
        connection: AutoKitteh connection name.

    Raises:
        ValueError: The connection name is invalid.
    """
    if not re.fullmatch(r"[A-Za-z_]\w*", connection):
        raise ValueError(f"Invalid AutoKitteh connection name: {connection!r}")


def encode_jwt(payload: dict[str, int], connection: str, algorithm: str) -> str:
    """Mock function to generate JWTs, overridden by the AutoKitteh runner."""
    print("!!!!!!!!!! SDK's encode_jwt() not overridden !!!!!!!!!!")
    return "DUMMY JWT"


def refresh_oauth(integration: str, connection: str) -> tuple[str, datetime]:
    """Mock function to refresh OAuth tokens, overridden by the AutoKitteh runner."""
    print("!!!!!!!!!! SDK's refresh_oauth() not overridden !!!!!!!!!!")
    return "DUMMY TOKEN", datetime.now(UTC).replace(tzinfo=None)



"""Initialize a Boto3 (AWS SDK) client, based on an AutoKitteh connection."""

import os

import boto3

from .connections import check_connection_name


def boto3_client(connection: str, service: str, region: str = "", **kwargs):
    """Initialize a Boto3 (AWS SDK) client, based on an AutoKitteh connection.

    API reference:
    https://boto3.amazonaws.com/v1/documentation/api/latest/index.html

    Code samples:
    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/examples.html

    Args:
        connection: AutoKitteh connection name.
        service: AWS service name.
        region: AWS region name.

    Returns:
        Boto3 client.

    Raises:
        ValueError: AutoKitteh connection or AWS service/region names are invalid.
        BotoCoreError: Authentication error.
    """
    check_connection_name(connection)

    if not service:
        raise ValueError("AWS service name is required")

    if not region:
        region = os.getenv(connection + "__Region", "")

    return boto3.client(
        service,
        region,
        aws_access_key_id=os.getenv(connection + "__AccessKeyID"),
        aws_secret_access_key=os.getenv(connection + "__SecretKey"),
        aws_session_token=os.getenv(connection + "__Token"),
        **kwargs,
    )



"""AutoKitteh Python SDK."""

from .attr_dict import AttrDict
from .decorators import activity, inhibit_activities
from .event import Event
from .events import next_event, subscribe, unsubscribe, start
from .signals import Signal, next_signal, signal
from . import errors


__all__ = [
    "AttrDict",
    "Event",
    "Signal",
    "activity",
    "errors",
    "inhibit_activities",
    "next_event",
    "next_signal",
    "signal",
    "start",
    "subscribe",
    "unsubscribe",
]



import os

import discord

from .connections import check_connection_name
from .errors import ConnectionInitError


def discord_client(connection: str, intents=None, **kwargs) -> discord.Client:
    """Initialize a Discord client, based on an AutoKitteh connection.

    API reference:
    https://discordpy.readthedocs.io/en/stable/api.html

    Args:
        connection: AutoKitteh connection name.
        intents: An object representing the events your bot can receive.

    Returns:
        Discord client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        DiscordException: Connection attempt failed, or connection is unauthorized.
    """
    check_connection_name(connection)
    if not intents:
        intents = discord.Intents.default()

    bot_token = os.getenv(connection + "__BotToken")
    if not bot_token:
        raise ConnectionInitError(connection)

    return discord.Client(intents=intents, **kwargs)


def bot_token(connection: str):
    check_connection_name(connection)

    bot_token = os.getenv(connection + "__BotToken")
    if not bot_token:
        raise ConnectionInitError(connection)

    return bot_token



"""Initialize Microsoft Graph SDK clients, based on AutoKitteh connections."""

from datetime import datetime, timedelta, UTC
import os

from azure.core import credentials
from azure import identity
from msgraph import GraphServiceClient

from .connections import check_connection_name, refresh_oauth
from .errors import ConnectionInitError


# Default buffer time to refresh OAuth tokens before they expire.
DEFAULT_REFRESH_BUFFER_TIME = timedelta(minutes=5)


def teams_client(connection: str, **kwargs) -> GraphServiceClient:
    """Initialize a Microsoft Teams client, based on an AutoKitteh connection.

    API documentation:
    https://docs.autokitteh.com/integrations/microsoft/teams/python

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Microsoft Graph client.

    Raises:
        ValueError: AutoKitteh connection name or auth type are invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        OAuthRefreshError: OAuth token refresh failed.
    """
    # TODO: Narrower scopes instead of all the default ones.
    return _microsoft_client(connection, **kwargs)


def _microsoft_client(connection: str, **kwargs) -> GraphServiceClient:
    """Initialize a Microsoft Graph client, based on an AutoKitteh connection.

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Microsoft Graph client.

    Raises:
        ValueError: AutoKitteh connection name or auth type are invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        OAuthRefreshError: OAuth token refresh failed.
    """
    check_connection_name(connection)

    auth_type = os.getenv(connection + "__auth_type", "")
    match auth_type:
        case "oauthDefault" | "oauthPrivate":
            creds = OAuthTokenProvider(connection)
        case "daemonApp":
            if os.getenv(connection + "__private_certificate"):
                creds = _certificate_credentials(connection)
            else:
                creds = _client_secret_credentials(connection)
        case _:
            err = f"AutoKitteh connection {connection!r} with "
            raise ValueError(err + f" invalid auth type {auth_type!r}")

    return GraphServiceClient(credentials=creds, **kwargs)


class OAuthTokenProvider(credentials.TokenCredential):
    """OAuth 2.0 token wrapper for Microsoft Graph clients."""

    def __init__(self, connection: str, buffer_time: timedelta | None = None) -> None:
        """Initialize credentials based on an AutoKitteh connection's variables.

        Used by server-default and private OAuth 2.0 user-delegated apps.

        Args:
            connection: AutoKitteh connection name.
            buffer_time: Buffer time to refresh OAuth tokens before they expire
                (optional, default = 5 minutes).

        Raises:
            ValueError: AutoKitteh connection name is invalid, or not using OAuth.
            ConnectionInitError: AutoKitteh connection was not initialized yet.
            OAuthRefreshError: OAuth token refresh failed.
        """
        self._connection = connection
        check_connection_name(connection)
        self._buffer_time = buffer_time or DEFAULT_REFRESH_BUFFER_TIME

        auth_type = os.getenv(connection + "__auth_type")
        if not auth_type:
            raise ConnectionInitError(connection)
        if not auth_type.startswith("oauth"):
            raise ValueError(f"AutoKitteh connection {connection!r} not using OAuth")

        self._access_token = os.getenv(connection + "__oauth_access_token", "")
        if not self._access_token:
            raise ConnectionInitError(connection)

        expiry = os.getenv(connection + "__oauth_expiry", "")
        self._expiry = datetime.fromisoformat(expiry).astimezone(UTC)
        self._refresh_token = os.getenv(connection + "__oauth_refresh_token", "")

    def get_token(self, *scopes: str, **kwargs) -> credentials.AccessToken:
        if self._expiry + self._buffer_time >= datetime.now(UTC):
            self._refresh_oauth_token()

        expires_on = int(self._expiry.timestamp())
        return credentials.AccessToken(token=self._access_token, expires_on=expires_on)

    def _refresh_oauth_token(self) -> None:
        self._access_token, self._expiry = refresh_oauth("microsoft", self._connection)
        self._expiry = self._expiry.astimezone(UTC)


def _certificate_credentials(connection: str) -> identity.CertificateCredential:
    """Initialize credentials based on an AutoKitteh connection's variables.

    Used by daemon (i.e. non-user-delegated) applications with advanced auth needs.

    Args:
        connection: AutoKitteh connection name.

    Raises:
        ValueError: AutoKitteh connection was configured to use OAuth
            (i.e. user-delegated permissions instead of application ones).
        ConnectionInitError: AutoKitteh connection was not initialized yet,
            or initialized to use a client secret instead of a certificate.
    """
    check_connection_name(connection)
    auth_type = os.getenv(connection + "__auth_type")
    if not auth_type:
        raise ConnectionInitError(connection)
    if auth_type.startswith("oauth"):
        raise ValueError(f"AutoKitteh connection {connection!r} using OAuth")

    client_id = os.getenv(connection + "__private_client_id")
    certificate = os.getenv(connection + "__private_certificate")
    if not client_id or not certificate:
        raise ConnectionInitError(connection)

    return identity.CertificateCredential(
        client_id=client_id,
        certificate_data=certificate.encode("utf-8"),
        tenant_id=os.getenv(connection + "__private_tenant_id", "") or "common",
    )


def _client_secret_credentials(connection: str) -> identity.ClientSecretCredential:
    """Initialize credentials based on an AutoKitteh connection's variables.

    Used by daemon (i.e. non-user-delegated) applications with simple auth needs.

    Args:
        connection: AutoKitteh connection name.

    Raises:
        ValueError: AutoKitteh connection was configured to use OAuth
            (i.e. user-delegated permissions instead of application ones).
        ConnectionInitError: AutoKitteh connection was not initialized yet.
    """
    check_connection_name(connection)
    auth_type = os.getenv(connection + "__auth_type")
    if not auth_type:
        raise ConnectionInitError(connection)
    if auth_type.startswith("oauth"):
        raise ValueError(f"AutoKitteh connection {connection!r} using OAuth")

    client_id = os.getenv(connection + "__private_client_id")
    client_secret = os.getenv(connection + "__private_client_secret")
    if not client_id or not client_secret:
        raise ConnectionInitError(connection)

    return identity.ClientSecretCredential(
        client_id=client_id,
        client_secret=client_secret,
        tenant_id=os.getenv(connection + "__private_tenant_id", "") or "common",
    )



"""AutoKitteh SDK errors."""

from google.auth.exceptions import RefreshError


class AutoKittehError(Exception):
    """Generic base class for all errors in the AutoKitteh SDK."""

    def __init__(self, *args):
        super().__init__(*args)


class ConnectionInitError(AutoKittehError):
    """A required AutoKitteh connection was not initialized yet."""

    def __init__(self, connection: str):
        super().__init__(f"AutoKitteh connection {connection!r} not initialized")


class EnvVarError(AutoKittehError):
    """A required environment variable is missing or invalid."""

    def __init__(self, env_var: str, desc: str):
        super().__init__(f"Environment variable {env_var!r} is {desc}")


class OAuthRefreshError(AutoKittehError):
    """OAuth token refresh failed."""

    def __init__(self, connection: str, error: RefreshError):
        super().__init__(f"OAuth refresh failed for {connection!r} connection: {error}")


class AtlassianOAuthError(AutoKittehError):
    """API calls not supported by OAuth-based Atlassian connections."""

    def __init__(self, connection: str):
        msg = f"API calls not supported by {connection!r}, "
        msg += "use a token-based connection instead"
        super().__init__(msg)



"""Send and receive signals."""

from datetime import timedelta
from dataclasses import dataclass

from autokitteh import AttrDict


@dataclass
class Signal:
    name: str
    payload: any = None


def signal(session_id: str, name: str, payload: any = None) -> None:
    """Signal a session."""
    # Dummy implementation for local development.
    pass


def next_signal(
    name: str | list[str], *, timeout: timedelta | int | float = None
) -> Signal | None:
    """Get the next signal."""
    # Dummy implementation for local development.
    return None



"""AutoKitteh Event class"""

from dataclasses import dataclass
from .attr_dict import AttrDict


@dataclass
class Event:
    """AutoKitteh Event as passed to entrypoints."""

    data: AttrDict
    session_id: str



#!/usr/bin/env python
"""Generate requirements from pyproject.toml"""

from argparse import ArgumentParser, FileType
import tomllib

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "input",
    help="input TOML",
    default="../pyproject.toml",
    type=FileType("rb"),
    nargs="?",
)
parser.add_argument(
    "output",
    help="output requirements file",
    default="requirements.txt",
    type=FileType("w"),
    nargs="?",
)
args = parser.parse_args()

header = f"""\
# Generated by gen-reqs.py, DO NOT EDIT
# This is an export of {args.input.name} for builds in readthedocs.
"""

config = tomllib.load(args.input)
print(header, file=args.output)
for dep in config["project"]["optional-dependencies"]["all"]:
    print(dep, file=args.output)



"""Initialize a Height client, based on an AutoKitteh connection."""

from requests import Session

from .oauth2_session import OAuth2Session


def height_client(connection: str) -> Session:
    """Initialize an Height client, based on an AutoKitteh connection.

    API reference:
    https://height-api.xyz/openapi/

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Requests session object.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
    """
    return OAuth2Session("height", connection)



"""Helper class to allow attribute access to dictionary keys."""


class AttrDict(dict):
    """Allow attribute access to dictionary keys.

    >>> config = AttrDict({'server': {'port': 8080}, 'debug': True})
    >>> config.server.port
    8080
    >>> config.debug
    True
    """

    def __getattr__(self, name: str):
        try:
            value = self[name]
            if isinstance(value, dict):
                value = AttrDict(value)
            return value
        except KeyError:
            # "from None" will remove the confusing KeyError from the stack trace
            raise AttributeError(name) from None

    def __setattr__(self, name: str, value):
        # The default __getattr__ doesn't fail but also don't change values.
        cls = self.__class__.__name__
        raise NotImplementedError(f"{cls} does not support setting attributes")

    def __dir__(self):
        return self.keys()



"""Slack client initialization, and other helper functions."""

import os
import re

from slack_sdk.web.client import WebClient

from .connections import check_connection_name
from .errors import ConnectionInitError


def slack_client(connection: str, **kwargs) -> WebClient:
    """Initialize a Slack client, based on an AutoKitteh connection.

    API reference:
    https://slack.dev/python-slack-sdk/api-docs/slack_sdk/web/client.html

    This function doesn't initialize a Socket Mode client because the
    AutoKitteh connection already has one to receive incoming events.

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Slack SDK client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        SlackApiError: Connection attempt failed, or connection is unauthorized.
    """
    check_connection_name(connection)

    bot_token = os.getenv(connection + "__oauth_access_token")  # OAuth v2
    if not bot_token:
        bot_token = os.getenv(connection + "__private_bot_token")  # Socket Mode
    if not bot_token:  # TODO(INT-267): Remove this old env var check.
        bot_token = os.getenv(connection + "__oauth_AccessToken")
    if not bot_token:  # TODO(INT-267): Remove this old env var check.
        bot_token = os.getenv(connection + "__BotToken")
    if not bot_token:
        raise ConnectionInitError(connection)

    client = WebClient(bot_token, **kwargs)
    client.auth_test().validate()
    return client


def normalize_channel_name(name: str) -> str:
    """Convert arbitrary text into a valid Slack channel name.

    See: https://api.slack.com/methods/conversations.create#naming

    Args:
        name: Desired name for a Slack channel.

    Returns:
        Valid Slack channel name.
    """
    if name == "":
        return name

    name = name.lower().strip()
    name = re.sub(r"['\"]", "", name)  # Remove quotes.
    name = re.sub(r"[^a-z0-9_-]", "-", name)  # Replace invalid characters.
    name = re.sub(r"[_-]{2,}", "-", name)  # Remove repeating separators.

    # Slack channel names are limited to 80 characters,
    # but that's too long for comfort, so we use 50 instead.
    name = name[:50]

    # Cosmetic tweak: remove leading and trailing hyphens.
    if name[0] == "-":
        name = name[1:]
    if name[-1] == "-":
        name = name[:-1]

    return name



import os

import asana

from .connections import check_connection_name
from .errors import ConnectionInitError


def asana_client(connection: str) -> asana.ApiClient:
    """Initialize an Asana client, based on an AutoKitteh connection.

    API reference:
    https://developers.asana.com/docs/python

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Asana client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
    """
    check_connection_name(connection)

    access_token = os.getenv(connection + "__pat")
    if not access_token:
        raise ConnectionInitError(connection)

    configuration = asana.Configuration()
    configuration.access_token = access_token
    return asana.ApiClient(configuration)



"""Initialize a Salesforce client, based on an AutoKitteh connection."""

import os

from simple_salesforce import Salesforce

from .connections import check_connection_name
from .errors import ConnectionInitError


def salesforce_client(connection: str, **kwargs) -> Salesforce:
    """Initialize a Salesforce client, based on an AutoKitteh connection.

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Salesforce client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        SalesforceApiError: Connection attempt failed, or connection is unauthorized.
    """
    check_connection_name(connection)

    token = os.getenv(connection + "__oauth_access_token")
    instance_url = os.getenv(connection + "__instance_url")
    if not token or not instance_url:
        raise ConnectionInitError(connection)

    return Salesforce(instance_url=instance_url, session_id=token, **kwargs)



import os

from .connections import check_connection_name
from .errors import ConnectionInitError

from auth0.management import Auth0


def auth0_client(connection: str, **kwargs) -> Auth0:
    """Initialize an Auth0 client, based on an AutoKitteh connection.

    API reference:
    https://auth0-python.readthedocs.io/en/latest/

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Auth0 SDK client.

    Raises:
        ConnectionInitError: If the connection is not initialized.
        ValueError: If the connection name is invalid.
    """
    check_connection_name(connection)

    token = os.getenv(connection + "__oauth_AccessToken")
    domain = os.getenv(connection + "__auth0_domain")
    if not token or not domain:
        raise ConnectionInitError(connection)

    return Auth0(domain, token, **kwargs)



import requests
from os import getenv

from .connections import refresh_oauth
from .connections import check_connection_name
from .errors import ConnectionInitError


class OAuth2Session(requests.Session):
    """Encapsulates arequests session, based on an AutoKitteh connection.

    - Automatically sets the Authorization header with an OAuth token.
    - Automatically refreshes an OAuth token if a refresh token is
      initialized in the connection.
    """

    def __init__(self, integration: str, connection: str, *args, **kwargs):
        """Initialize a requests session, based on an AutoKitteh connection.

        Args:
            connection: AutoKitteh integration and connection name.

        Returns:
            OAuth2Session.

        Raises:
            ValueError: AutoKitteh connection name is invalid.
            ConnectionInitError: AutoKitteh connection was not initialized yet.
        """
        check_connection_name(connection)

        super().__init__(*args, **kwargs)

        self._integration = integration
        self._connection = connection
        self._refresh_token = getenv(f"{connection}__oauth_refresh_token")
        self._token = getenv(f"{connection}__oauth_access_token")

        if not self._token:
            raise ConnectionInitError(connection)

        self._set_token(self._token)

        self.hooks["response"].append(self._intercept)

    def _intercept(self, r, *args, **kwargs):
        if not self._refresh_token or r.status_code != 401:
            return r

        token, _ = refresh_oauth(self._integration, self._connection)
        self._set_token(token)
        r.request.headers["Authorization"] = f"Bearer {token}"

        return self.send(r.request)

    def _set_token(self, token):
        self._token = token
        self.headers.update({"Authorization": f"Bearer {token}"})



"""Initialize a GitHub client, based on an AutoKitteh connection."""

import os
import time
from urllib.parse import urljoin

from github import Auth, Consts, Github, GithubIntegration

from .connections import check_connection_name, encode_jwt
from .errors import ConnectionInitError


def github_client(connection: str, **kwargs) -> Github:
    """Initialize a GitHub client, based on an AutoKitteh connection.

    API reference and examples: https://pygithub.readthedocs.io/

    Args:
        connection: AutoKitteh connection name.

    Returns:
        PyGithub client.

    Raises:
        ValueError: AutoKitteh connection name or GitHub app IDs are invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
    """
    check_connection_name(connection)

    # Optional: GitHub Enterprise Server
    base_url = os.getenv("GITHUB_ENTERPRISE_URL")
    if base_url:
        kwargs["base_url"] = urljoin(base_url, "api/v3")
        print("GitHub Enterprise base URL: " + kwargs["base_url"])

    # PAT + webhook
    pat = os.getenv(f"{connection}__pat")
    if pat:
        return Github(auth=Auth.Token(pat), **kwargs)

    # GitHub App (JWT)
    app_id = os.getenv(f"{connection}__app_id")
    if not app_id:
        raise ConnectionInitError(connection)

    install_id = os.getenv(f"{connection}__install_id")
    if not install_id:
        raise ConnectionInitError(connection)

    app = None
    private_key = os.getenv(f"{connection}__private_key")
    if private_key:
        # Exposing the private key is fine here as it belongs to the user.
        app = GithubIntegration(auth=Auth.AppAuth(int(app_id), private_key), **kwargs)
    else:
        app = GithubIntegration(auth=AppAuth(int(app_id), connection), **kwargs)
    return app.get_github_for_installation(int(install_id))


class AppAuth(Auth.AppAuth):
    """Generate JWTs without exposing the GitHub app's private key.

    Based on: https://github.com/PyGithub/PyGithub/blob/main/github/Auth.py
    """

    def __init__(self, app_id: int, ak_connection_name: str):
        self._app_id = app_id
        self._private_key = ak_connection_name
        self._jwt_expiry = Consts.DEFAULT_JWT_EXPIRY
        self._jwt_issued_at = Consts.DEFAULT_JWT_ISSUED_AT
        self._jwt_algorithm = Consts.DEFAULT_JWT_ALGORITHM

    def create_jwt(self, expiration: int | None = None) -> str:
        now = int(time.time())
        payload = {
            "iat": now + self._jwt_issued_at,
            "exp": now + (expiration if expiration is not None else self._jwt_expiry),
            "iss": self._app_id,
        }
        # This is the only change from the original code: replace the call to jwt.encode().
        # We don't monkey-patch it because the jwt module is usable outside the GitHub client too.
        return encode_jwt(payload, self.private_key, self._jwt_algorithm)



"""Initialize Google API clients, based on AutoKitteh connections."""

from datetime import UTC, datetime
import json
import os
import re

from google.auth.exceptions import RefreshError
from google.auth.transport.requests import Request
import google.generativeai as genai
import google.oauth2.credentials as credentials
import google.oauth2.service_account as service_account
from googleapiclient.discovery import build

from .connections import check_connection_name, refresh_oauth
from .errors import ConnectionInitError, OAuthRefreshError


def gmail_client(connection: str, **kwargs):
    """Initialize a Gmail client, based on an AutoKitteh connection.

    API documentation:
    https://docs.autokitteh.com/integrations/google/gmail/python

    Code samples:
    - https://github.com/autokitteh/kittehub/tree/main/samples/google/gmail
    - https://github.com/googleworkspace/python-samples/tree/main/gmail

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Gmail client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        OAuthRefreshError: OAuth token refresh failed.
    """
    # https://developers.google.com/gmail/api/auth/scopes
    default_scopes = [
        "https://www.googleapis.com/auth/gmail.modify",
        "https://www.googleapis.com/auth/gmail.settings.basic",
    ]
    creds = google_creds("gmail", connection, default_scopes, **kwargs)
    return build("gmail", "v1", credentials=creds, **kwargs)


def google_calendar_client(connection: str, **kwargs):
    """Initialize a Google Calendar client, based on an AutoKitteh connection.

    API documentation:
    https://docs.autokitteh.com/integrations/google/calendar/python

    Code samples:
    https://github.com/autokitteh/kittehub/tree/main/samples/google/calendar

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Google Calendar client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        OAuthRefreshError: OAuth token refresh failed.
    """
    # https://developers.google.com/calendar/api/auth
    default_scopes = [
        "https://www.googleapis.com/auth/calendar",
        "https://www.googleapis.com/auth/calendar.events",
    ]
    creds = google_creds("googlecalendar", connection, default_scopes, **kwargs)
    return build("calendar", "v3", credentials=creds, **kwargs)


def google_drive_client(connection: str, **kwargs):
    """Initialize a Google Drive client, based on an AutoKitteh connection.

    API documentation:
    https://docs.autokitteh.com/integrations/google/drive/python

    Code samples:
    https://github.com/googleworkspace/python-samples/tree/main/drive

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Google Drive client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        OAuthRefreshError: OAuth token refresh failed.
    """
    # https://developers.google.com/drive/api/guides/api-specific-auth
    default_scopes = [
        "https://www.googleapis.com/auth/drive.file",  # See ENG-1701
    ]
    creds = google_creds("googledrive", connection, default_scopes, **kwargs)
    return build("drive", "v3", credentials=creds, **kwargs)


def google_forms_client(connection: str, **kwargs):
    """Initialize a Google Forms client, based on an AutoKitteh connection.

    API documentation:
    https://docs.autokitteh.com/integrations/google/forms/python

    Code samples:
    - https://github.com/autokitteh/kittehub/tree/main/samples/google/forms
    - https://github.com/googleworkspace/python-samples/tree/main/forms

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Google Forms client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        OAuthRefreshError: OAuth token refresh failed.
    """
    # https://developers.google.com/identity/protocols/oauth2/scopes#script
    default_scopes = [
        "https://www.googleapis.com/auth/forms.body",
        "https://www.googleapis.com/auth/forms.responses.readonly",
    ]
    creds = google_creds("googleforms", connection, default_scopes, **kwargs)
    return build("forms", "v1", credentials=creds, **kwargs)


def gemini_client(connection: str, **kwargs) -> genai.GenerativeModel:
    """Initialize a Gemini generative AI client, based on an AutoKitteh connection.

    API reference:
    - https://ai.google.dev/gemini-api/docs
    - https://github.com/google-gemini/generative-ai-python/blob/main/docs/api/google/generativeai/GenerativeModel.md

    Code samples:
    - https://ai.google.dev/gemini-api/docs#explore-the-api
    - https://ai.google.dev/gemini-api/docs/text-generation?lang=python
    - https://github.com/google-gemini/generative-ai-python/tree/main/samples
    - https://github.com/google-gemini/cookbook

    Args:
        connection: AutoKitteh connection name.

    Returns:
        An initialized GenerativeModel instance.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
    """
    check_connection_name(connection)

    # Set the API key, if possible.
    api_key = os.getenv(connection + "__api_key")
    if not api_key:
        raise ConnectionInitError(connection)

    genai.configure(api_key=api_key)
    return genai.GenerativeModel(**kwargs)


def google_sheets_client(connection: str, **kwargs):
    """Initialize a Google Sheets client, based on an AutoKitteh connection.

    API documentation:
    https://docs.autokitteh.com/integrations/google/sheets/python

    Code samples:
    - https://github.com/autokitteh/kittehub/tree/main/samples/google/sheets
    - https://github.com/googleworkspace/python-samples/tree/main/sheets

    Args:
        connection: AutoKitteh connection name.

    Returns:
        Google Sheets client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        OAuthRefreshError: OAuth token refresh failed.
    """
    # https://developers.google.com/sheets/api/scopes
    default_scopes = ["https://www.googleapis.com/auth/spreadsheets"]
    creds = google_creds("googlesheets", connection, default_scopes, **kwargs)
    return build("sheets", "v4", credentials=creds, **kwargs)


def google_creds(integration: str, connection: str, scopes: list[str], **kwargs):
    """Initialize credentials for a Google APIs client, for service discovery.

    This function supports both AutoKitteh connection modes:
    users (with OAuth 2.0), and GCP service accounts (with a JSON key).

    Code samples:
    https://github.com/googleworkspace/python-samples

    For subsequent usage details, see:
    https://googleapis.github.io/google-api-python-client/docs/epy/googleapiclient.discovery-module.html#build

    Args:
        integration: AutoKitteh integration name.
        connection: AutoKitteh connection name.
        scopes: List of OAuth permission scopes.

    Returns:
        Google API credentials, ready for usage
        in "googleapiclient.discovery.build()".

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        OAuthRefreshError: OAuth token refresh failed.
    """
    check_connection_name(connection)

    if os.getenv(connection + "__authType") == "oauth":  # User (OAuth 2.0)
        return _google_creds_oauth2(integration, connection, scopes)

    json_key = os.getenv(connection + "__JSON")  # Service Account (JSON key)
    if json_key:
        # https://google-auth.readthedocs.io/en/stable/reference/google.oauth2.service_account.html#google.oauth2.service_account.Credentials.from_service_account_info
        return service_account.Credentials.from_service_account_info(
            json.loads(json_key), scopes=scopes, **kwargs
        )

    raise ConnectionInitError(connection)


def _google_creds_oauth2(integration: str, connection: str, scopes: list[str]):
    """Initialize user credentials for Google APIs using OAuth 2.0.

    For more details, see:
    - https://google-auth.readthedocs.io/en/stable/reference/google.oauth2.credentials.html#google.oauth2.credentials.Credentials.from_authorized_user_info
    - https://github.com/googleapis/google-auth-library-python/blob/main/google/oauth2/credentials.py

    Args:
        integration: AutoKitteh integration name.
        connection: AutoKitteh connection name.
        scopes: List of OAuth permission scopes.

    Returns:
        Google API credentials, ready for usage
        in "googleapiclient.discovery.build()".

    Raises:
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        OAuthRefreshError: OAuth token refresh failed.
    """
    expiry = os.getenv(connection + "__oauth_Expiry")
    if not expiry:
        raise ConnectionInitError(connection)

    # Convert Go's time string (e.g. "2024-06-20 19:18:17 -0700 PDT") to
    # an ISO-8601 string that Python can parse with timezone awareness.
    expiry = re.sub(r" [A-Z]+.*", "", expiry)
    expiry = re.sub(r"\.\d+", "", expiry)  # Also ignore sub-second precision.
    dt = datetime.fromisoformat(expiry).astimezone(UTC).replace(tzinfo=None)

    token = os.getenv(connection + "__oauth_AccessToken")
    client_secret = os.getenv("GOOGLE_CLIENT_SECRET", "NOT AVAILABLE")

    if client_secret == "NOT AVAILABLE":
        # In Docker/Cloud environments, handle token refresh through AutoKitteh
        creds = credentials.Credentials(token=token, expiry=dt, scopes=scopes)
        creds.refresh_handler = _google_refresh_handler(integration, connection)
    else:
        # Refreshes to be handled by the Python client, as usual.
        creds = credentials.Credentials.from_authorized_user_info(
            {
                "token": token,
                "refresh_token": os.getenv(connection + "__oauth_RefreshToken"),
                "expiry": dt.isoformat(),
                "client_id": os.getenv("GOOGLE_CLIENT_ID"),
                "client_secret": client_secret,
                "scopes": scopes,
            }
        )

    try:
        if creds.expired:
            creds.refresh(Request())
    except RefreshError as e:
        raise OAuthRefreshError(connection, e)

    return creds


def _google_refresh_handler(integration: str, connection: str) -> callable:
    """Refresh handler for OAuth 2.0 user credentials for Google APIs.

    For more details, see:
    - https://google-auth.readthedocs.io/en/stable/reference/google.oauth2.credentials.html
    - https://github.com/googleapis/google-auth-library-python/blob/main/google/oauth2/credentials.py

    Args:
        integration: AutoKitteh integration name.
        connection: AutoKitteh connection name.

    Returns:
        Generated function (based on the input) to return a fresh access token
        and its expiry date. Overridden by AutoKitteh to keep the Google client
        secret hidden from workflows, and unused in local development.
    """

    def impl(request, scopes: list[str]) -> tuple[str, datetime]:
        return refresh_oauth(integration, connection)

    return impl


def google_id(url: str) -> str:
    """Extract the Google Doc/Form/Sheet ID from a URL. This function is idempotent.

    Example: 'https://docs.google.com/.../d/1a2b3c4d5e6f/edit' --> '1a2b3c4d5e6f'
    """
    match = re.match(r"(.*/d/(e/)?)?([\w-]{20,})", url)
    if match:
        return match.group(3)
    else:
        raise ValueError(f'Invalid Google ID in "{url}"')



"""Decorator to mark a function as a Temporal activity."""

ACTIVITY_ATTR = "__ak_activity__"
INHIBIT_ACTIVITIES_ATTR = "__ak_inhibit_activities__"


def activity(fn: callable) -> callable:
    """Decorator to mark a function as a Temporal activity.

    This forces AutoKitteh to run the function as a single Temporal activity,
    instead of a sequence of activities within a Temporal workflow.

    Use this decorator when you want to run functions with input arguments
    and/or return values that are not compatible with pickle.

    For more details, see:
    https://docs.python.org/3/library/pickle.html#what-can-be-pickled-and-unpickled
    """
    setattr(fn, ACTIVITY_ATTR, True)
    return fn


def inhibit_activities(fn: callable) -> callable:
    """Decorator to inhibit the execution of Temporal activities.

    Functions using this decorator will not spawn activities even if otherwise
    activities should have been launched. This is useful for performing operations
    that are required to run even on replay (such as various clients creation) and
    are completely deterministic. The function results are not cached and would be
    rerun in case of replay.

    CAVEAT: Do not use this on functions that take a long time to run (more than a second),
    as they will cause the workflow to timeout.
    """
    setattr(fn, INHIBIT_ACTIVITIES_ATTR, True)
    return fn



"""Initialize a HubSpot client, based on an AutoKitteh connection."""

import os

from hubspot import HubSpot

from .connections import check_connection_name, refresh_oauth
from .errors import ConnectionInitError, OAuthRefreshError


def hubspot_client(connection: str, **kwargs) -> HubSpot:
    """Initialize a HubSpot client, based on an AutoKitteh connection.

    Args:
        connection: AutoKitteh connection name.

    Returns:
        HubSpot SDK client.

    Raises:
        ValueError: AutoKitteh connection name is invalid.
        ConnectionInitError: AutoKitteh connection was not initialized yet.
        OAuthRefreshError: OAuth token refresh failed.
    """
    check_connection_name(connection)

    refresh_token = os.getenv(connection + "__oauth_RefreshToken")
    if not refresh_token:
        raise ConnectionInitError(connection)

    try:
        access_token, _ = refresh_oauth("hubspot", connection)
    except Exception as e:
        raise OAuthRefreshError(connection, str(e)) from e

    return HubSpot(access_token=access_token, **kwargs)
